#! /usr/bin/env node
var fs            = require('fs');
var _cwd          = require('cwd');
global.cwd        = _cwd();
var isRoot        = require('is-root');
var program       = require('commander');
var Table         = require('cli-table');
var prompt        = require('co-prompt');
var co            = require('co');
var ncp           = require('ncp').ncp;
var colors        = require('colors');
var builder      = require('./lib/builder.js');
var _builder;

program
.arguments('<command>')
.option('--i-suck', 'you should really figure out how to chown this directory instead of running sudo')
.option('-p, --port <port>', 'port to attach to')
.option('-h, --host <host>', 'host address')
.option('-F, --show-file', 'show file of routes controller')
.option('-f, --file <file>', 'file to test')
.action(function(command){
  switch(command) {
    case 'init':
       init(program.iSuck);
       break;
    case 'server':
      findCurrentWorkingDirectory();
      _builder = builder.build();
      startServers(program.port, program.host);
      break;
    case 'routes':
      findCurrentWorkingDirectory();
      _builder = builder.build();
      displayRoutes(program.showFile);
      break;
    case 'test':
      findCurrentWorkingDirectory();
      _builder = builder.build();
      //perform_tests(program.file);
      break;
    default:
      console.error("no such command %s", command);
  }
})
.parse(process.argv);

function init(iSuck) {
  if (isRoot() && !iSuck) {
     console.error("running `$ sudo intuition init` is not a good idea. If you really want to ignore this advice please add the flag --i-suck");
  } else {
    var intuitionDirectoryPath = cwd + '/.intuition';
    if (fs.existsSync(intuitionDirectoryPath)) {
      console.error('An intuition project already exists in ' + cwd);
      process.exit();
    }

    var intuitionDependenciesPath = __dirname + '/copy';
    console.log("Press ^C at any time to quit.");
      co(function*() {
        var defaultProjectName = cwd.split('/' )[cwd.split('/' ).length-1];
        var projectName = yield prompt('project-name: (' + defaultProjectName + ') ');
        var defaultVersion = "1.0.0";
        var version = yield prompt('version: (' + defaultVersion + ') ');
        var author = yield prompt('author: ');
        var manifest = {
          "project-name": projectName || defaultProjectName,
          "version": version || defaultVersion,
          "author":  author
        };

        ncp(intuitionDependenciesPath, cwd, function(err) {
          if (err) {
            console.error(err);
          } else {
            fs.writeFileSync(cwd + '/.intuition/manifest.json', JSON.stringify(manifest), 'utf-8');
            console.log(colors.green('successfully created ' + manifest["project-name"]));
            console.log("now setup your database by configuring", "config/database.json".cyan)
            process.exit(1);
          }
        });

      })
  }
};

function startServers(port, host) {

  _builder.startDatabase()
  .then(function(db){
      _builder.startServerWithDatabase(port, host, db);
   });

}

function displayRoutes(showfile) {

  if(showfile) {
    var table = new Table({
        head: ['URI', 'HTTP METHOD', 'CONTROLLER', 'METHOD', 'FILE'],
        colWidths: [25, 14, 35, 25, 50]
    });

    var routesArray = global.Router.routesToArray().reverse();
    routesArray.forEach(function(route) {
      var uri = route.route;
      var httpMethods = route.http;
      for(let method in httpMethods) {
        var controller =   httpMethods[method].controller;
        var controllerMethod = httpMethods[method].method;
        for (var i = 0; i < _builder.applicationList.length; ++i) {
          if (_builder.applicationList[i].namespaceString === controller) {
            var file = _builder.applicationList[i].filename;
            table.push([uri, method, controller, controllerMethod, file.replace(__dirname + '/', '')]);
            break;
          }
        };
      }
    });
  }  else {
    var table = new Table({
        head: ['URI', 'HTTP METHOD', 'CONTROLLER', 'METHOD'],
        colWidths: [25, 14, 35, 25]
    });

    var routesArray = global.Router.routesToArray().reverse();
    routesArray.forEach(function(route) {
      var uri = route.route;
      var httpMethods = route.http;
      for(let method in httpMethods) {
        var controller       = httpMethods[method].controller;
        var controllerMethod = httpMethods[method].method;
        table.push([uri, method, controller, controllerMethod]);
      }
    });
  }
  console.log(table.toString());
}
function findCurrentWorkingDirectory() {
  var intuitionDirectoryPath = cwd + '/.intuition';
  var cwdArray = cwd.split("/");
  if(cwdArray.length === 0) {
    console.error("No intuition project detected. Could not start servers");
    process.exit();
  }
  if (!fs.existsSync(intuitionDirectoryPath)) {
    cwdArray.pop();
    global.cwd = cwdArray.join("/");
    findCurrentWorkingDirectory();
  }
}




function perform_tests(filename) {

//  global.test           = require( './$test' ).test;
//  global.skip           = require( './$test' ).skip;
//  global.assert         = require( './$test' ).assert;
//  global.assertEquals   = require( './$test' ).assertEquals;
//  global.assertFalse    = require( './$test' ).assertFalse;
//
//  var Tester = function() {
//
//    this.count         = 0;
//    this.successCount  = 0;
//    this.failCount     = 0;
//    this.errorCount    = 0;
//    this.skipCount     = 0;
//  }
//
//  var tester = new Tester();
//  global.test = test.bind(tester);
//
//  if (filename) {
//
//    require(__dirname + '/tests/' + filename);
//
//  } else {
//
//    var tests = _builder.traverseTests();
//    tests.forEach(function(test){
//       require(test.filename);
//    });
//
//  }
//
//  console.error({Tests: tester.count,
//                 Successes: tester.successCount,
//                 Fails: tester.failCount,
//                 Errors: tester.errorCount,
//                 Skips: tester.skipCount,
//                });
//  process.exit()
};
