var MongoDB            = require('mongodb');
var MongoClient        = MongoDB.MongoClient;
var Grid               = require('gridfs-stream');
var colors             = require('colors');
var session            = require('express-session');
var MongoDBStore       = require('connect-mongodb-session')(session);
var cwd                = global.cwd;

var GridObject = function(buffer, options) {
  if (!this.buffer) throw new Error('No buffer in GridObject');
  this.options = this.options || {};
}

global.GridObject = GridObject;

var MonogoDB = function() {
  var self = this;

  try {
  this.dataTypes = {
      'double'              : MongoDB.Double,
      'string'              : String,
      'object'              : Object,
      'array'               : Array,
      'binary'              : MongoDB.Binary,
      'bool'                : Boolean,
      'date'                : Date,
      'timestamp'           : MongoDB.timestamp,
      'regex'               : RegExp,
      'DBPointer'           : MongoDB.DBRef,
      'symbol'              : MongoDB.Symbol,
      'integer'             : MongoDB.Long,
      'grid'                : GridObject
    };
  } catch (e) {
    console.log(e)
  }

  this.connection = new Promise(function(resolve, reject) {
    var _config;

    try {
      var databaseJsonPath = global.cwd + '/config/database.json';
      _config = require(databaseJsonPath);
    } catch (e) {
      console.error("there was a problem parsing config/database.json" , e)
    }

    if (!_config.host ||
        _config.host.length == 0 ||
        !_config.port ||
        _config.port.length == 0 ||
        !_config.name ||
        _config.name.length == 0 ||
        !_config.username ||
        _config.username.length == 0 ||
        !_config.password ||
        _config.password.length == 0) {
      console.error("missing config vars for db connection config/database.js")
      process.exit();
    }
    var dbConnectString = 'mongodb://' + _config.username + ':' + _config.password + '@' + _config.host + ':' + _config.port + '/' + _config.name;

    MongoClient.connect(dbConnectString, function(error, db) {

      if(!error) {
        console.log("database connected ...".cyan);
        db.dataTypes = self.dataTypes;
        self.db = db;
        var GridFileSystem = Grid(db, MongoDB);
        resolve({
                  database: self.db,
                  store: new MongoDBStore({
                          uri: dbConnectString,
                          grid: GridFileSystem,
                          collection: 'Sessions'
                  })
                });

      } else{
        console.log(error);
        reject(error);
      }

    });
  })
};

var dbs = {
  'MonogoDB' : MonogoDB
}

module.exports = MonogoDB;