namespace('Models')
var ObjectID      = require('mongodb').ObjectID;
var getCallerFile = require('./../imports' ).getCallerFile;
var co            = require('co');

var Model = Intuition.extend(function() {

  this.__collection = null;
  this.__oneToOneReferences = {};
  this.__oneToManyReferences = {};
  this.__relationships = {};
  this.__relations = {};
  this.__accessors = {};
  this.__joined = {};
  this.__generators = [];
  this.__disabledCoroutines = [];
  this.__generatedAccessors = [];
  this.__enableCoroutines = true;

  this.__typeProperties = {};
  this.__required = [];
  this.hidden = [];

  this.createdAt;
  this.updatedAt;
  this.autoCreatedAtTimestamp = true;
  this.autoUpdatedAtTimestamp = true;
  this.__callbacks = {};


  Object.defineProperty(this, 'disabledCoroutines', {
    get: function() {
      return this.__disabledCoroutines;
    },
    set: function(values) {
      this.__disabledCoroutines = this.__disabledCoroutines.concat(values);
    },
    configurable: true
  });

  Object.defineProperty(this, 'properties', {
    set: function(_properties) {
      for(let _property in _properties) {
        this.__typeProperties[_property] = _properties[_property]
      }
    },
    get: function() {
      return this.__typeProperties;
    },
    configurable: true
  })

  return {

    init: function(args) {
      this._setTableName();
      if (!Model.tables[this.__table]) Model.tables[this.__table] = this.__class;
      this.__collection = _database.database.collection(this.__table);

      if (this.autoCreatedAtTimestamp) this.properties['createdAt'] = 'date';
      if (this.autoUpdatedAtTimestamp) this.properties['updatedAt'] = 'date';

      this.initProperties(args);

    },

    initProperties: function(args) {

      this.__properties = args || {};
      this._id = this.__properties._id;

      Object.defineProperty(this, 'id', {
        get: function() {
          return this.__properties._id;
        },
        configurable: true
      });

      Object.defineProperty(this, 'required', {
        get: function() {
          return this.__required;
        },
        set: function(_required) {
          this.__required = this.__required.concat(_required);
        },
        configurable: true
      });

      for(let property in args) {
        if(args[property]._bsontype !== 'DBRef' && !this.__relations[property]) {

          Object.defineProperty(this, property, {
            get: function() {
              return this.__properties[property];
            },
            set: function(value) {
               this.__properties[property] = value;
            },
            configurable: true
          });

        } else if (this.__oneToOneReferences[property]) {
           var __id = args[property].oid;
           var Class = this.__oneToOneReferences[property].class;
           var accessor = this.__oneToOneReferences[property].accessor;
           this.__relationships[accessor] = __id;
        } else if (this.__oneToManyReferences[property]) {
          var __ids = args[property];
          var Class = this.__oneToManyReferences[property].class
          var accessor = this.__oneToManyReferences[property].accessor;
          this.__relationships[accessor] = __ids;
        }
      }
      return this;
    },

    _setTableName: function() {
      this.__table = this.__class__.replace(/([A-Z])/g, function($1){return "_"+$1.toLowerCase();}).substring(1) + 's';
    },

    update: function() {

      for(var i = 0; i < this.required.length; ++i) {
         if(!this[this.required[i]]) {
           throw new ModelPropertyException(this.table + ' missing required property ' + this.required[i]);
           return;
         }
      }

      if (this.__callbacks.beforeUpdate) this.__callbacks.beforeUpdate.call(this);

      if (this.autoCreatedAtTimestamp) this.createdAt = new Date();
      if (this.autoUpdatedAtTimestamp) this.updatedAt = new Date();

      var document = {};

      for(var _property in this.properties) {

        var value = this[_property];
        var _type = this.properties[_property];
        var Type = _database.database.dataTypes[_type];

        if(!(document[_property] instanceof Type)) {
          document[_property] = new Type(value);
        } else {
          document[_property] = value;
        }

        if( document[_property] instanceof String) {
          document[_property] = document[_property].valueOf();
        }
      }

      var updateMany = {};
      var hasRelations = false;
      for(var _relation in this.__relations) {
        if(this.__oneToOneReferences[_relation]) {
          document[_relation] = new _database.database.dataTypes.DBPointer(this.__relations[_relation].__table,
                                                                  this.__relations[_relation].id);
        } else if (this.__oneToManyReferences[_relation]) {
          var ids = this.__relations[_relation]._ids;
          updateMany[_relation] = { $each : ids};
          hasRelations = true;
        }
      }
      if(hasRelations) {
        var subDocument = {};
        subDocument.$addToSet = updateMany;
        subDocument['upsert'] = true;
        this.__collection.update( {_id : this._id}, subDocument);

      }

      if(this.autoUpdatedAtTimestamp) {
        document['updatedAt'] = new Date();
      }

      document['upsert'] = true;

      this.__collection.update( {_id : this._id}, {$set : document});

      delete document.upsert;

      for(var property in document) {
        if(property !== '$addToSet') this.__properties[property] = document[property];
      }

      if (this.__callbacks.afterUpdate) this.__callbacks.afterUpdate.call(this);
    },

    save: function() {

      for(var i = 0; i < this.required.length; ++i) {
         if(!this[this.required[i]]) {
           throw new ModelPropertyException(this.table + ' missing required property ' + this.required[i]);
           return;
         }
      }

      var document = {};

      for(var _property in this.properties) {

        var value = this[_property];
        var _type = this.properties[_property];
        var Type = _database.database.dataTypes[_type];

        if(_property !== 'createdAt' && _property !== 'updatedAt') {
          document[_property] = new Type(value);
          if( document[_property].valueOf() !== undefined ) {
            document[_property] = document[_property].valueOf();
          }
        }
      }

      for(var _relation in this.__relations) {
        if(this.__oneToOneReferences[_relation]) {
          document[_relation] = new _database.database.dataTypes.DBPointer(this.__relations[_relation].__table,
                                                                  this.__relations[_relation].id);
        } else if (this.__oneToManyReferences[_relation]) {
          document[_relation] = this.__relations[_relation]._ids;
        }
      }

      if(this.autoCreatedAtTimestamp) {
        document['createdAt'] = new Date();
      }
      if(this.autoUpdatedAtTimestamp) {
        document['updatedAt'] = new Date();
      }


      this.__collection.insert(document);
      for(var property in document) {
        this.__properties[property] = document[property];
      }
    },

    _delete: function() {

      if (this.__callbacks.beforeDelete) this.__callbacks.beforeDelete.call(this);
      this.__table.remove({_id : this.id}, true)
      if (this.__callbacks.afterDelete) this.__callbacks.afterDelete.call(this);

    },

    beforeDelete: function(callback) {
      this.__callbacks.beforeDelete = callback;
    },

    afterDelete: function(callback) {
      this.__callbacks.afterDelete = callback;
    },

    beforeUpdate: function(callback) {
      this.__callbacks.beforeUpdate = callback;
    },

    afterUpdate: function(callback) {
      this.__callbacks.afterUpdate = callback;
    },

    //must add dependency to required queue
    belongsTo: function(Class, alias) {

      var accessor = '$' + alias
      this.__oneToOneReferences[alias] = {class: Class, accessor: accessor};

      Object.defineProperty(this, alias, {
        set: function(value) {
          this.__relations[alias] = value;
          this.__joined[accessor] = value;
        },
        get : function() {
          try {
            return this.__joined[accessor];
          } catch (e) {
            return undefined;
          }
        },
        configurable: true
      })

      var accessorFunction = function *(resolve, reject) {

        if (this.__joined[accessor]) {
          resolve(this.__joined[accessor]);
        } else {
          try {
            var modelInstance = yield Class.$get(this.__relationships[accessor]);
            resolve( modelInstance );
         } catch (e) {
            reject(e);
         }
        }
      }

      this[accessor] = function() {
        return this.$coroutine(accessorFunction.bind(this));
      }.bind(this);

      this.__accessors[alias] = 'belongs';
    },

    hasOne: function(Class, alias, foreignAlias) {
      var accessor = '$' + alias;

      this.__oneToOneReferences[alias] = {class: Class, accessor: accessor, foreignAlias: foreignAlias};

      //creates the ability to set the object
      Object.defineProperty(this, alias, {
        set: function(value) {
          this.__joined[accessor] = value;
        },
        get : function() {
          try {
            return this.__relations[alias].id;
          } catch (e) {
            return undefined;
          }
        },
        configurable: true
      })

      var accessorFunction = function *(resolve, reject) {

        if (this.__joined[accessor]) {
          resolve(this.__joined[accessor]);
        } else {
          try {
		        var query = {};
		        var queryString = foreignAlias + '.$id';
		        query[queryString] = this.id;
            var modelInstance = yield Class.$whereOne(query);
            resolve( modelInstance );
         } catch (e) {
            reject(e);
         }
        }
     }
     this[accessor] = function() {
       return this.$coroutine(accessorFunction.bind(this));
     }.bind(this);

     this.__accessors[alias] = 'has';
    },

    belongsToMany: function(Class, alias) {
      var accessor = '$' + alias;
      var setter = alias;

      this.__relations[setter] = ModelCollection.factory(Class);
      this.__oneToManyReferences[alias] = {class: Class, accessor: accessor};

      Object.defineProperty(this, alias, {
        set: function(value) {
          for(var i = 0; i < value.length; ++i) this.__relations[alias].push(value[i]);
        },
        get : function() {
          return this.__relations[alias];
        },
        configurable: true
      });

      var accessorFunction = function *(resolve, reject) {

        if (this.__joined[accessor]) {
          resolve(this.__joined[accessor]);
        } else {
          try {
            var collection = yield Class.$in('_id', this.__relationships[accessor]);
            resolve( collection );
         } catch (e) {
            reject(e);
         }
        }
      }

      this[accessor] = function() {
        return this.$coroutine(accessorFunction.bind(this));
      }.bind(this);

      this.__accessors[alias] = 'belongsMany';
    },


    hasMany: function(Class, alias, foreignAlias) {
      var accessor = '$' + alias

      this.__oneToManyReferences[alias]  = {class: Class, accessor: accessor, foreignAlias: foreignAlias};

      Object.defineProperty(this, alias, {
        set: function(value) {
          this.__joined[accessor] = value;
        },
        get : function() {
          try {
            return this.__joined[accessor]
          } catch (e) {
            return undefined;
          }
        },
        configurable: true
      });

      var accessorFunction = function *(resolve, reject) {

        if (this.__joined[accessor]) {
          resolve(this.__joined[accessor]);
        } else {
          try {
            var pseudoInstanceOfRelations = Class.new();
            var query = {};
            if (pseudoInstanceOfRelations.__oneToOneReferences[foreignAlias]) {
              var queryString = foreignAlias + '.$id';
              query[queryString] = this.id;
            } else if (pseudoInstanceOfRelations.__oneToManyReferences[foreignAlias]) {
              query[foreignAlias] = {$in: [this.id]}
            }
            var modelCollection = yield Class.$where(query);
            resolve( modelCollection )

         } catch (e) {
            reject(e);
         }
        }

     }
     this[accessor] = function() {
       return this.$coroutine(accessorFunction.bind(this));
     }.bind(this);


      this.__accessors[alias] = 'hasMany'
    },

    _$getOne: function(__id) {
      return this.$coroutine(function* (resolve, reject) {
        try {
          var model = yield this.__collection.findOne({_id : __id});
          resolve(model);
        } catch (error) {
          reject(error);
        }
      });
    }
  }

});

Model.instance = function() {

  if(!this._instance) {
    this._instance = this.new();
  }
  return this._instance;
}


Model.$all = function() {
  return this.$where({});
}

Model.$where = function(query) {
  var self = this;
  var instance = this.instance();
  return this.$coroutine(function *(resolve, reject){
    try {
      var q = instance.__collection.find(query);
      var modelCollection = yield ModelCollection.new(q, self);
      resolve(modelCollection);
    } catch (error) {
      reject(error);
    }
  });
}

Model.$whereOne = function(query) {
		return this.$coroutine(function *(resolve, reject){
			try {
					var modelCollection = yield this.$where(query);
					resolve(modelCollection[0]);
			}	catch (error) {
				reject(error)
			}
		});
}

Model.$in = function(property, array) {
  var self = this;
  var instance = this.instance();
  return this.$coroutine(function *(resolve, reject){
    try {
      var query = {};
      query[property] = {'$in' : array};
      var q = instance.__collection.find(query);
      var modelCollection = yield ModelCollection.new(q, self);
      resolve(modelCollection);
    } catch (error) {
      reject(error);
    }
  });
}


Model.__getByID = function(id) {
  var instance = this.instance();
  return instance.__collection.findOne(ObjectID(id));
}

Model.$get = function(id) {
  var _Model = this;
  var instance = this.instance();
  return this.$coroutine(function*(resolve, reject) {
      try {
        var model = yield instance._$getOne(ObjectID(id));
        var instace = _Model.new(model)
        resolve(instace);
      } catch (error) {
        reject(error);
      }
  });
}

Model.tables = {};

var ModelPropertyException = function(message) {
  this.name = 'ModelPropertyException';
  this.message = message;
}

ModelPropertyException.prototype = new Error;


module.exports = Model;