namespace('Models')

var ModelCollection = Intuition.extend(function() {
  Array.call(this);
  this._ids = [];

  return {
    init: function(query, Model) {
      if (query &&  Model) {
        var pseudoInstance = Model.new();
        this._pseudoInstance = pseudoInstance;
        this.Model = Model;

        var promise = this.$coroutine(function*(resolve, reject){
            var results = yield query.toArray();
            setTimeout(function() {
              try {
                results.forEach(function(model) {
                  var copy = new Model;
                  copy.initProperties(model);
                  this.push(copy);
                }.bind(this));
                resolve(this);
              } catch (error) {
                reject(error)
              }
            }.bind(this));
        });
        return promise;
      } else {
        //not sure where else this will not be the case but okay
      }

    },

    push: function(modelInstance) {
      if (!(modelInstance instanceof this.Model)) {
        throw new ModelPropertyException('CANNOT PUSH ' + modelInstance.__class__ + ' is not of type ' + this.Model.__class__);
      } else {
        this._ids.push(modelInstance.__properties._id);
        Array.prototype.push.call(this, modelInstance);
      }
    },

    concat: function(modelCollection) {
      for(var i = 0; i < modelCollection.length; ++i) {
        this.push(modelCollection[i]);
      }
    },

    $join: function (alias) {
        if(this._pseudoInstance.__accessors[alias] === 'has') {
          return this.__has(alias);
        } else if (this._pseudoInstance.__accessors[alias] === 'belongs') {
          return this.__belongs(alias);
        } else if (this._pseudoInstance.__accessors[alias] === 'belongsMany') {
          return this.__belongsMany(alias);
        } else if (this._pseudoInstance.__accessors[alias] === 'hasMany') {
          return this.__hasMany(alias);
        }
    },

    __has: function(alias) {
      return this.$coroutine(function *(resolve, reject) {
        var Model = this._pseudoInstance.__oneToOneReferences[alias].class;
        var foreignAlias = this._pseudoInstance.__oneToOneReferences[alias].foreignAlias;
        var accessor = this._pseudoInstance.__oneToOneReferences[alias].accessor;
        var relations = yield Model.$in(foreignAlias + '.$id', this._ids);
        setTimeout(function() {
          var joinMap = {};
          for(var i = 0; i < relations.length; ++i) {
            joinMap[relations[i].__properties[foreignAlias].oid.id] =  relations[i];
          }

          for(var i = 0; i < this.length; ++i) {
            var key = this[i]._id.id
            var relation = joinMap[key];
            this[i].__joined[accessor] = relation;
          }
          resolve(this)
        }.bind(this));
      });

    },

    __belongs: function(alias) {
      return this.$coroutine(function *(resolve, reject) {
        var Model = this._pseudoInstance.__oneToOneReferences[alias].class;
        var accessor = this._pseudoInstance.__oneToOneReferences[alias].accessor;
        var relationIds = [];

        yield this.$coroutine(function*(resolve) {
          setTimeout(function(){
            this.forEach(function(model) {
              relationIds.push(model.__properties[alias].oid);
            });
            resolve();
          }.bind(this));
        })
        var relations = yield Model.$in('_id', relationIds);
        setTimeout(function() {

          var joinMap = {};
          for(var i = 0; i < relations.length; ++i) {
            joinMap[relations[i]._id.id] =  relations[i];
          }

          for(var i = 0; i < this.length; ++i) {
            var key = this[i].__properties[alias].oid.id;
            var relation = joinMap[key];
            this[i].__joined[accessor] = relation;
          }
          resolve(this)
        }.bind(this));
      });
    },

    __belongsMany: function(alias) {
      return this.$coroutine(function*(resolve, reject) {
        var Model = this._pseudoInstance.__oneToManyReferences[alias].class;
        var accessor = this._pseudoInstance.__oneToManyReferences[alias].accessor;
        var relationIds = yield this._fetchRelations(alias);
        var relations = yield Model.$in("_id", relationIds);
        var relationsMap = yield this.$coroutine(function*(resolve, reject) {
          setTimeout(function() {
            var map = {}
            relations.forEach(function ( model ) {
                map[ JSON.stringify(model.__properties._id) ] = model;
            }, {});
              resolve(map);
          }.bind(this));
        });

         setTimeout(function() {
          for(var i = 0; i < this.length; i++) {
            this[i].__joined[accessor] = ModelCollection.factory(Model);
            for(var j = 0; j < this[i].__properties[alias].length; ++j) {
              var id = this[i].__properties[alias][j];
              var modelInstance = relationsMap[JSON.stringify(id)];
              this[i].__joined[accessor].push(modelInstance);
            }
          }
          resolve(this);
         }.bind(this));
      });
    },

    _fetchRelations: function (property) {
      return this.$coroutine(function*(resolve, reject) {
        setTimeout(function(){
          var relations = [];
          this.forEach(function(model){
            if(model.__properties[property]) relations = relations.concat(model.__properties[property]);
          });
          resolve(relations);
        }.bind(this))
      });
    },

    __hasMany: function(alias) {
      return this.$coroutine(function*(resolve, reject) {
        //is it a one to Many relationship?
        var Model = this._pseudoInstance.__oneToManyReferences[alias].class; //Like class
        var foreignAlias = this._pseudoInstance.__oneToManyReferences[alias].foreignAlias;  //comment
        var accessor = this._pseudoInstance.__oneToManyReferences[alias].accessor;
        var relationPseudoInstance = Model.new();
        //is it a one to Many or a many to many?
        if(relationPseudoInstance.__oneToOneReferences[foreignAlias]) {
          // ONE TO MANY
          var relations = yield Model.$in(foreignAlias + '.$id', this._ids)
          var relationMap = yield this.$coroutine(function *(resolve, reject) {
             try {
               var map = {};
               setTimeout(function() {
                 relations.forEach(function(model){
                    var key = model.__properties[foreignAlias].oid.id;
                    if (!map[key]) map[key] = [];
                    map[key].push(model);
                 });
                 resolve(map);
               });
             } catch (e) {
               reject(e);
             }
          });
          this.forEach(function(model) {
             var relation = relationMap[model._id.id];
             model.__joined[alias] = relation;
          });
          resolve(this);

        } else {
          // MANY TO MANY
          var query = {};
          query[foreignAlias] = {$in: this._ids}
          var relations = yield Model.$where(query);
          var selfRelationsMap = yield this.$coroutine(function*(resolve, reject) {
            setTimeout(function() {
              var map = {}
              this.forEach(function ( model ) {
                  map[ JSON.stringify(model.__properties._id) ] = model;
              }, {});
                resolve(map);
            }.bind(this));
          });

          setTimeout(function() {
            for(var i = 0; i < relations.length; ++i) {
              var selfRelations = relations[i].__properties[foreignAlias];
              for(var j =0; j < selfRelations.length; ++j) {
                var key = JSON.stringify(selfRelations[j]);
                if(!this._selfHash()[key].__joined[alias]) {
                  this._selfHash()[key].__joined[alias] = ModelCollection.factory(Model);
                }
                this._selfHash()[key].__joined[alias].push(relations[i])
              }
            }
            resolve(this);
          }.bind(this));
        }
      });
    },

    _selfHash: function() {
      if(!this.selfAsHash) {
        this.selfAsHash = {};
        for(var i = 0; i < this.length; i++) {
          this.selfAsHash[JSON.stringify(this[i].id)] = this[i];
        }
      }
      return this.selfAsHash;
    },

    $get: function(id) {
      //retunr one
    },

    $where: function(query) {
      return this.$coroutine(function *(resolve, reject) {
        query[_id] = {$in: this._ids}
        resolve(yield this.Model.$where());
      });
    },

    /**
     * empty will break all references if an owner is specified
     * */
    empty: function () {

    },

    emptyWhere: function(query) {

    },

    $update: function(properties) {
      return this.$coroutine(function *(resolve, reject) {
        for(var i = 0; i < this.length; ++i ){
          var model = this[i];
          for(let _property in properties) {
            model[_property] = properties[_property];
          }
          model.update();
        }
        resolve(this);
      });
    },

    $updateWhere: function(properties, query) {
      return this.$coroutine(function *(resolve, reject) {
        var models = yield this.$where(query);
        resolve(yield models.$update(properties));
      });
    },

    $delete: function(properties) {
      return this.$coroutine(function *(resolve, reject) {
        for(var i = 0; i < this.length; ++i ){
          for(let _property in properties) {
            this[i][_property] = properties[_property];
          }
          this[i].delete();
        }
        resolve(this);
      });
    },

    $deleteWhere: function (properties, query) {
      return this.$coroutine(function *(resolve, reject) {
        var models = yield this.$where(query);
        resolve(yield models.$delete(properties));
      });
    }

  }
});

ModelCollection.prototype = Object.create(Array.prototype);

ModelCollection.factory = function(Class) {
  var _instance = new ModelCollection;
  _instance.Model = Class;
  _instance.__class__ = 'ModelCollection';
  return _instance;
}

module.exports = ModelCollection;