namespace('Views')
var dot          = require('dot');
dot.templateSettings = require(cwd + '/config/dot.js');
var sanitizeHTML = require('sanitize-html');

var View = Intuition.extend(function() {

  this.args = [];
  this.template = null;

  this.__generators = [];
  this.__disabledCoroutines = [];
  this.__json = null;
  this.__enableCoroutines = true;

  Object.defineProperty(this, 'disabledCoroutines', {
    get: function() {
      return this.__disabledCoroutines;
    },
    set: function(values) {
      this.__disabledCoroutines = this.__disabledCoroutines.concat(values);
    },
    configurable: true
  });

  return {

    init: function(_args) {
      for (var i = 0; i < this.args.length; ++i) {
        this[this.args[i]] = _args[this.args[i]];
        if (!this[this.args[i]]) {
          throw new Error("Missing Argument" + this.args[i] + " for " + this.__class__);
        }
      }
      if (this.template) this._template = dot.template(this.template);

      return this.$coroutine( function *(resolve, reject) {
        try {
          yield this.$render();
          resolve(this);
        } catch (e) {
          reject(e);
        }
      });
    },

    $json: function *(resolve) {
      resolve();
    },

    $render: function *(resolve, reject) {
      try {
        this.__json = yield this.$json();
        this.toString = function() {
          if (this.template) {
            for(let prop in this.__json) {
              if (this.__json[prop] instanceof String) {
                this.__json[prop] = sanitizeHTML(this.__json[prop]);
              }
            }
            return this._template(this.__json);
          } else {
            return this.renderJSON();
          }
        }
        resolve(this.__json);
      } catch (e) {
        reject(e);
      }
    },

    renderJSON: function (resolve, reject) {
      return JSON.stringify(this);
    },

    renderTemplate: function (resolve, reject) {
      return this.toString();
    },

  }

});

View.prototype.toJSON = function() {
  return this.__json;
};


module.exports = View;