var Router = function() {}

Router.routes = {};

Router.routesToArray = function() {
   var routeArray = [];

   for(let route in this.routes) {
      routeArray.push({http:this.routes[route], route: route});
   }

   routeArray.sort(function(a,b) {

   var binaryA = Router.getBinaryValue( a.route );
   var binaryB = Router.getBinaryValue( b.route );

   if (binaryA > binaryB) {
       return -1;
     } else if (binaryB > binaryA) {
       return 1;
     } else {
       return 0;
     }
   });

  return routeArray;
}

Router.getBinaryValue = function(route) {

  var uriArr = route.split('/');
  var binaryArray = [];
  if (route === '/' || route === '') return -1;
  uriArr.forEach(function(component) {
    if ( component !== '') {
      if (component.substr(0,1) === ":") {
        binaryArray.push(0);
      } else {
        binaryArray.push(1);
      }
    }
  });
  let binaryString = binaryArray.join('');
  return parseInt(binaryString, 2);
}


Router.declare = function(namespace, routeObject) {
  var routes = this._processRoutes(routeObject);
  for(let route in routes) {
    if(route[0] !== '/') {
      console.error('route: "' + route + '" in ' + namespace +' is missing "/", please change to /' + route)
      process.exit()
    }
    if(!this.routes[route]) this.routes[route] = {}
    for(let http in routes[route]) {
      this.routes[route][http] = {
        controller: namespace,
        method: routes[route][http].method
      }
    }
  }

}

Router._processRoutes = function(routeObject, namespace) {
  var routes = {};
  if(typeof routeObject === 'string') {
    //if(routeObject.slice(-1) !== '/') routeObject += '/';

    routes[routeObject] = {
      'GET' : {
        controller : namespace,
        method     : 'index',
      },
      'POST' : {
        controller : namespace,
        method     : 'create',
      },
    };

    if(routeObject == '/') routeObject = '';
    routes[routeObject + '/:var'] = {
      'GET' : {
        controller : namespace,
        method     : 'show',
      },
      'PUT' : {
        controller : namespace,
        method     : 'updated',
      },
      'PATCH' : {
        controller : namespace,
        method     : 'updatedAttribute',
      },
      'DELETE'     : {
        controller : namespace,
        method     : 'updatedAttribute',
      }
    };

    routes[routeObject + '/:var/new'] = {
      'GET' : {
        controller : namespace,
        method     : 'new',
      }
    };

    routes[routeObject + '/:var/edit'] = {
      'GET' : {
        controller : namespace,
        method     : 'edit',
      },
    };
    return routes;
  }

  for(let route in routeObject) {
    //is it undefined? then defined it
    if(!routes[route]) routes[route] = {};

    for(let http in routeObject[route]) {
      routes[route][http] = {
        controller : namespace,
        method: routeObject[route][http]
      };
    }
  }

  return routes;
}

module.exports = Router;